/*---------------------------------------------------------------------------*/

/*!  \file  celllist.c
     \brief  create or update the linked cell list.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/

/*  DO NOT EDIT THIS FILE. This file is generated from celllist.c.m4.  */

#include "common.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "xmalloc.h"
#include "vector.h"
#include "celllist.h"

#define R_CUT 2.0

/*---------------------------------------------------------------------------*/
static linked_cell_list *allocate_linked_cell_list2(soft_particles2 *particles,
                                                    box_geometry *geometry);
static void clear_linked_cell_list2(linked_cell_list *celllist,
                                    soft_particles2 *particles);
static void construct_linked_cell_list2(linked_cell_list *celllist,
                                        vector2 **r,
                                        soft_particles2 *particles,
                                        box_geometry *geometry);

static linked_cell_list *allocate_linked_cell_list3(soft_particles3 *particles,
                                                    box_geometry *geometry);
static void clear_linked_cell_list3(linked_cell_list *celllist,
                                    soft_particles3 *particles);
static void construct_linked_cell_list3(linked_cell_list *celllist,
                                        vector3 **r,
                                        soft_particles3 *particles,
                                        box_geometry *geometry);

/*---------------------------------------------------------------------------*/
linked_cell_list *allocate_linked_cell_list(soft_particles *particles,
                                            box_geometry *geometry)
{
    linked_cell_list *celllist;

        switch(particles->dimension)
    {
        case 2:
            celllist = allocate_linked_cell_list2(&(particles->particles2),geometry);
            break;

        case 3:
            celllist = allocate_linked_cell_list3(&(particles->particles3),geometry);
            break;

        default:
            celllist = NULL;
            break;
    }

    return celllist;
}

/*---------------------------------------------------------------------------*/
void free_linked_cell_list(linked_cell_list *celllist)
{
    if(celllist == NULL)
        return;

    free(celllist->head[0]);
    free(celllist->list[0]);
    free(celllist->head);
    free(celllist->list);
    free(celllist);
}

/*---------------------------------------------------------------------------*/
void construct_linked_cell_list(linked_cell_list *celllist,
                                soft_particles *particles,
                                box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            construct_linked_cell_list2(celllist,particles->particles2.r,
                                                  &(particles->particles2),geometry);
            break;

        case 3:
            construct_linked_cell_list3(celllist,particles->particles3.r,
                                                  &(particles->particles3),geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void construct_linked_cell_list_star(linked_cell_list *celllist,
                                     soft_particles *particles,
                                     box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            construct_linked_cell_list2(celllist,particles->particles2.r_star,
                                                  &(particles->particles2),geometry);
            break;

        case 3:
            construct_linked_cell_list3(celllist,particles->particles3.r_star,
                                                  &(particles->particles3),geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
linked_cell_list *allocate_linked_cell_list2(soft_particles2 *particles,
                                               box_geometry *geometry)
{
    linked_cell_list *celllist;
    int i,nc,m_tot;

    celllist = xmalloc(sizeof(linked_cell_list));

    celllist->dimension = 2;
    celllist->nx = (int)floor(geometry->l[0] / R_CUT);
    if(celllist->nx <= 0)
        celllist->nx = 1;
    celllist->ny = (int)floor(geometry->l[1] / R_CUT);
    if(celllist->ny <= 0)
        celllist->ny = 1;
    celllist->nz = 1;

    celllist->head = xmalloc(sizeof(int *) * particles->n);
    celllist->list = xmalloc(sizeof(int *) * particles->n);

    nc = celllist->nx * celllist->ny;

    m_tot = total_number_of_soft_particles2(particles);

    celllist->head[0] = xmalloc(sizeof(int) * nc * particles->n);
    celllist->list[0] = xmalloc(sizeof(int) * m_tot);

    for(i = 0;i < particles->n;i++)
    {
        if(i > 0)
        {
            celllist->head[i] = celllist->head[i - 1] + nc;
            celllist->list[i] = celllist->list[i - 1] + particles->m[i - 1];
        }
    }

    clear_linked_cell_list2(celllist,particles);

    return celllist;
}

/*---------------------------------------------------------------------------*/
void clear_linked_cell_list2(linked_cell_list *celllist,
                               soft_particles2 *particles)
{
    int i,j,nc;

    nc = celllist->nx * celllist->ny;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < nc;j++)
        {
            celllist->head[i][j] = -1;
        }
        for(j = 0;j < particles->m[i];j++)
        {
            celllist->list[i][j] = -1;
        }
    }
}

/*---------------------------------------------------------------------------*/
void construct_linked_cell_list2(linked_cell_list *celllist,
                                 vector2 **r,
                                 soft_particles2 *particles,
                                 box_geometry *geometry)
{
    int i,j,nc;
    int index[2];
    int cell_index;
    vector2 c;

    c[0] = celllist->nx / geometry->l[0];
    c[1] = celllist->ny / geometry->l[1];
    nc = celllist->nx * celllist->ny;

    clear_linked_cell_list2(celllist,particles);

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
                /*  calculate the index for the cell.  */
            index[0] = (int)(r[i][j][0] * c[0]);
index[1] = (int)(r[i][j][1] * c[1]);

            cell_index = index[1] * celllist->nx + index[0];

                /*  add the particle index into the list.  */
            celllist->list[i][j] = celllist->head[i][cell_index];
            celllist->head[i][cell_index] = j;
        }
    }
}

/*---------------------------------------------------------------------------*/
linked_cell_list *allocate_linked_cell_list3(soft_particles3 *particles,
                                               box_geometry *geometry)
{
    linked_cell_list *celllist;
    int i,nc,m_tot;

    celllist = xmalloc(sizeof(linked_cell_list));

    celllist->dimension = 3;
    celllist->nx = (int)floor(geometry->l[0] / R_CUT);
    if(celllist->nx <= 0)
        celllist->nx = 1;
    celllist->ny = (int)floor(geometry->l[1] / R_CUT);
    if(celllist->ny <= 0)
        celllist->ny = 1;
    celllist->nz = (int)floor(geometry->l[2] / R_CUT);
           if(celllist->nz <= 0)
               celllist->nz = 1;

    celllist->head = xmalloc(sizeof(int *) * particles->n);
    celllist->list = xmalloc(sizeof(int *) * particles->n);

    nc = celllist->nx * celllist->ny * celllist->nz;

    m_tot = total_number_of_soft_particles3(particles);

    celllist->head[0] = xmalloc(sizeof(int) * nc * particles->n);
    celllist->list[0] = xmalloc(sizeof(int) * m_tot);

    for(i = 0;i < particles->n;i++)
    {
        if(i > 0)
        {
            celllist->head[i] = celllist->head[i - 1] + nc;
            celllist->list[i] = celllist->list[i - 1] + particles->m[i - 1];
        }
    }

    clear_linked_cell_list3(celllist,particles);

    return celllist;
}

/*---------------------------------------------------------------------------*/
void clear_linked_cell_list3(linked_cell_list *celllist,
                               soft_particles3 *particles)
{
    int i,j,nc;

    nc = celllist->nx * celllist->ny * celllist->nz;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < nc;j++)
        {
            celllist->head[i][j] = -1;
        }
        for(j = 0;j < particles->m[i];j++)
        {
            celllist->list[i][j] = -1;
        }
    }
}

/*---------------------------------------------------------------------------*/
void construct_linked_cell_list3(linked_cell_list *celllist,
                                 vector3 **r,
                                 soft_particles3 *particles,
                                 box_geometry *geometry)
{
    int i,j,nc;
    int index[3];
    int cell_index;
    vector3 c;

    c[0] = celllist->nx / geometry->l[0];
    c[1] = celllist->ny / geometry->l[1];
    c[2] = celllist->nz / geometry->l[2];
           nc = celllist->nx * celllist->ny * celllist->nz;

    clear_linked_cell_list3(celllist,particles);

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
                /*  calculate the index for the cell.  */
            index[0] = (int)(r[i][j][0] * c[0]);
index[1] = (int)(r[i][j][1] * c[1]);
index[2] = (int)(r[i][j][2] * c[2]);

            cell_index = (index[2] * celllist->ny + index[1]) * celllist->nx + index[0];

                /*  add the particle index into the list.  */
            celllist->list[i][j] = celllist->head[i][cell_index];
            celllist->head[i][cell_index] = j;
        }
    }
}

/*---------------------------------------------------------------------------*/
