/*---------------------------------------------------------------------------*/

/*!  \file  force.c
     \brief  calculate the random and interaction forces.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2007,2008 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/

/*  DO NOT EDIT THIS FILE. This file is generated from force.c.m4.  */

#include "common.h"

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "vector.h"
#include "particles.h"
#include "nrand.h"
#include "force.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#ifndef M_SQRTPI
#define M_SQRTPI 1.77245385090551602729
#endif

#define SQR(x) ((x)*(x))

#define INDEX(i,n) (((i) == -1) ? ((n) - 1) : (((i) == (n)) ? (0) : (i)))
#define LATTICE_VECTOR(i,n,c) (((i) == -1) ? (c) : (((i) == (n)) ? (-(c)) : (0)))
#define IS_LOWER_BOUNDARY(i1,i2,n) (((i1) == 0) && ((i2) == -1))
#define IS_UPPER_BOUNDARY(i1,i2,n) (((i1) == ((n) - 1)) && ((i2) == (n)))

/*---------------------------------------------------------------------------*/
static void generate_random_force2(soft_particles2 *particles);
static void clear_interaction_force2(soft_particles2 *particles);
static void clear_stress_tensor2(soft_particles2 *particles);

static void generate_random_force3(soft_particles3 *particles);
static void clear_interaction_force3(soft_particles3 *particles);
static void clear_stress_tensor3(soft_particles3 *particles);

static void accumulate_tethering_interaction_stress(soft_particles *particles,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_stress(soft_particles *particles,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);


static void accumulate_tethering_interaction_force2(vector2 **r,vector2 **f,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_force2(vector2 **r,vector2 **f,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);
static void accumulate_site_interaction_force2_in_different_cells(
    vector2 **r,vector2 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector2 *c);
static void accumulate_site_interaction_force2_in_same_cell(
    vector2 **r,vector2 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector2 *c);

static void accumulate_tethering_interaction_force3(vector3 **r,vector3 **f,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_force3(vector3 **r,vector3 **f,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);
static void accumulate_site_interaction_force3_in_different_cells(
    vector3 **r,vector3 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector3 *c);
static void accumulate_site_interaction_force3_in_same_cell(
    vector3 **r,vector3 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector3 *c);

static void accumulate_tethering_interaction_stress2(vector2 **r,symmetric_tensor2 **sigma,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_stress2(vector2 **r,symmetric_tensor2 **sigma,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);
static void accumulate_site_interaction_stress2_in_different_cells(
    vector2 **r,symmetric_tensor2 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector2 *c);
static void accumulate_site_interaction_stress2_in_same_cell(
    vector2 **r,symmetric_tensor2 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector2 *c);

static void accumulate_tethering_interaction_stress3(vector3 **r,symmetric_tensor3 **sigma,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_stress3(vector3 **r,symmetric_tensor3 **sigma,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);
static void accumulate_site_interaction_stress3_in_different_cells(
    vector3 **r,symmetric_tensor3 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector3 *c);
static void accumulate_site_interaction_stress3_in_same_cell(
    vector3 **r,symmetric_tensor3 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector3 *c);

/*---------------------------------------------------------------------------*/
void generate_random_force(soft_particles *particles)
{
        switch(particles->dimension)
    {
        case 2:
            generate_random_force2(&(particles->particles2));
            break;

        case 3:
            generate_random_force3(&(particles->particles3));
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void clear_interaction_force(soft_particles *particles)
{
        switch(particles->dimension)
    {
        case 2:
            clear_interaction_force2(&(particles->particles2));
            break;

        case 3:
            clear_interaction_force3(&(particles->particles3));
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void clear_stress_tensor(soft_particles *particles)
{
        switch(particles->dimension)
    {
        case 2:
            clear_stress_tensor2(&(particles->particles2));
            break;

        case 3:
            clear_stress_tensor3(&(particles->particles3));
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_force(soft_particles *particles,
                                            polymer_blend *blend,
                                            box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            accumulate_tethering_interaction_force2(particles->particles2.r,
                        particles->particles2.f,blend,geometry);
            break;

        case 3:
            accumulate_tethering_interaction_force3(particles->particles3.r,
                        particles->particles3.f,blend,geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_force_star(soft_particles *particles,
                                                 polymer_blend *blend,
                                                 box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            accumulate_tethering_interaction_force2(particles->particles2.r_star,
                        particles->particles2.f_star,blend,geometry);
            break;

        case 3:
            accumulate_tethering_interaction_force3(particles->particles3.r_star,
                        particles->particles3.f_star,blend,geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force(soft_particles *particles,
                                       linked_cell_list *celllist,
                                       polymer_blend *blend,
                                       box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            accumulate_site_interaction_force2(particles->particles2.r,
                                                       particles->particles2.f,
                                                       celllist,blend,geometry);
            break;

        case 3:
            accumulate_site_interaction_force3(particles->particles3.r,
                                                       particles->particles3.f,
                                                       celllist,blend,geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force_star(soft_particles *particles,
                                            linked_cell_list *celllist,
                                            polymer_blend *blend,
                                            box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            accumulate_site_interaction_force2(particles->particles2.r_star,
                                                       particles->particles2.f_star,
                                                       celllist,blend,geometry);
            break;

        case 3:
            accumulate_site_interaction_force3(particles->particles3.r_star,
                                                       particles->particles3.f_star,
                                                       celllist,blend,geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_stress(soft_particles *particles,
                                            polymer_blend *blend,
                                            box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            accumulate_tethering_interaction_stress2(particles->particles2.r,
                        particles->particles2.sigma,blend,geometry);
            break;

        case 3:
            accumulate_tethering_interaction_stress3(particles->particles3.r,
                        particles->particles3.sigma,blend,geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress(soft_particles *particles,
                                       linked_cell_list *celllist,
                                       polymer_blend *blend,
                                       box_geometry *geometry)
{
        switch(particles->dimension)
    {
        case 2:
            accumulate_site_interaction_stress2(particles->particles2.r,
                                                       particles->particles2.sigma,
                                                       celllist,blend,geometry);
            break;

        case 3:
            accumulate_site_interaction_stress3(particles->particles3.r,
                                                       particles->particles3.sigma,
                                                       celllist,blend,geometry);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_interaction_force_without_noise(soft_particles *particles,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    clear_interaction_force(particles);

    construct_linked_cell_list(celllist,particles,geometry);

    accumulate_site_interaction_force(particles,celllist,blend,geometry);
    accumulate_tethering_interaction_force(particles,blend,geometry);
}

/*---------------------------------------------------------------------------*/
void calculate_stress_tensor(soft_particles *particles,
                             linked_cell_list *celllist,
                             polymer_blend *blend,
                             box_geometry *geometry)
{
    clear_stress_tensor(particles);

    construct_linked_cell_list(celllist,particles,geometry);

    accumulate_site_interaction_stress(particles,celllist,blend,geometry);
    accumulate_tethering_interaction_stress(particles,blend,geometry);
}

/*---------------------------------------------------------------------------*/
void generate_random_force2(soft_particles2 *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            particles->w[i][j][0] = nrand();
particles->w[i][j][1] = nrand();
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_interaction_force2(soft_particles2 *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            particles->f[i][j][0] = 0.0;
particles->f[i][j][1] = 0.0;

            particles->f_star[i][j][0] = 0.0;
particles->f_star[i][j][1] = 0.0;
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_stress_tensor2(soft_particles2 *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            particles->sigma[i][j][0] = 0.0;
particles->sigma[i][j][1] = 0.0;
particles->sigma[i][j][2] = 0.0;
        }
    }
}

/*---------------------------------------------------------------------------*/
void generate_random_force3(soft_particles3 *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            particles->w[i][j][0] = nrand();
particles->w[i][j][1] = nrand();
particles->w[i][j][2] = nrand();
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_interaction_force3(soft_particles3 *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            particles->f[i][j][0] = 0.0;
particles->f[i][j][1] = 0.0;
particles->f[i][j][2] = 0.0;

            particles->f_star[i][j][0] = 0.0;
particles->f_star[i][j][1] = 0.0;
particles->f_star[i][j][2] = 0.0;
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_stress_tensor3(soft_particles3 *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            particles->sigma[i][j][0] = 0.0;
particles->sigma[i][j][1] = 0.0;
particles->sigma[i][j][2] = 0.0;
particles->sigma[i][j][3] = 0.0;
particles->sigma[i][j][4] = 0.0;
particles->sigma[i][j][5] = 0.0;
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_force2(vector2 **r,vector2 **f,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    int i,j,k;
    vector2 dr;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];

                if(dr[0] >= 0.5 * geometry->l[0])
                          {
                              dr[0] -= geometry->l[0];
                          }
                          else if(dr[0] < - 0.5 * geometry->l[0])
                          {
                              dr[0] += geometry->l[0];
                          }
if(dr[1] >= 0.5 * geometry->l[1])
                          {
                              dr[1] -= geometry->l[1];
                          }
                          else if(dr[1] < - 0.5 * geometry->l[1])
                          {
                              dr[1] += geometry->l[1];
                          }

                f[i][j][0] -= dr[0];
                          f[k][j][0] += dr[0];
f[i][j][1] -= dr[1];
                          f[k][j][1] += dr[1];
            }
        }
    }
    else
    {
            /*  box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];

                if(dr[1] >= 0.5 * geometry->l[1])
                {
                    dr[1] -= geometry->l[1];
                    dr[0] -= geometry->delta_l;
                }
                else if(dr[1] < - 0.5 * geometry->l[1])
                {
                    dr[1] += geometry->l[1];
                    dr[0] += geometry->delta_l;
                }
                while(dr[0] >= 0.5 * geometry->l[0])
                {
                    dr[0] -= geometry->l[0];
                }
                while(dr[0] < - 0.5 * geometry->l[0])
                {
                    dr[0] += geometry->l[0];
                }
                

                f[i][j][0] -= dr[0];
                          f[k][j][0] += dr[0];
f[i][j][1] -= dr[1];
                          f[k][j][1] += dr[1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force2(vector2 **r,vector2 **f,
                                          linked_cell_list *celllist,
                                          polymer_blend *blend,
                                          box_geometry *geometry)
{
    int i1,j1,i2,j2,i2_i,j2_i;
    int cell_index1,cell_index2;
    vector2 c;
    int i2_min,i2_max;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        
        for(j1 = 0;j1 < celllist->ny;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = j1 * celllist->nx + i1;

                
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = j2_i * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_force2_in_same_cell(
                                r,f,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_force2_in_different_cells(
                                r,f,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                

            }
        }
        
    }
    else
    {
            /*  box deformation  */
        

            /*  boundary cells (j = 0, ny - 1)  */
        for(j1 = 0;j1 < celllist->ny;j1 += (celllist->ny - 1))
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = j1 * celllist->nx + i1;

                
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else
                    {
                        i2_min = i1 - 1;
                        i2_max = i1 + 1;
                    }

                    for(i2 = i2_min;i2 <= i2_max;i2++)
                    {
                        i2_i = i2;
                        c[0] = 0.0;

                        while(i2_i < 0)
                        {
                            i2_i += celllist->nx;
                            c[0] += geometry->l[0];
                        }
                        while(i2_i >= celllist->nx)
                        {
                            i2_i -= celllist->nx;
                            c[0] -= geometry->l[0];
                        }

                        if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] += geometry->delta_l;
                        }
                        else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] -= geometry->delta_l;
                        }


                        cell_index2 = j2_i * celllist->nx + i2_i;

                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_force2_in_same_cell(
                                r,f,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_force2_in_different_cells(
                                r,f,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                
            }
        }


            /*  non-boundary cells  */
        for(j1 = 1;j1 < celllist->ny - 1;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = j1 * celllist->nx + i1;

                
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = j2_i * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_force2_in_same_cell(
                                r,f,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_force2_in_different_cells(
                                r,f,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                

            }
        }
        
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force2_in_different_cells(
    vector2 **r,vector2 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector2 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector2 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
        for(i2 = 0;i2 < blend->n;i2++)
        {
            epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k1];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k2];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];

                    dr2 = SQR(dr[0]) + SQR(dr[1]);

                    f[i1][j1][0] += epsilon * exp(-dr2) * 2.0 * dr[0];
                              f[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0];
f[i1][j1][1] += epsilon * exp(-dr2) * 2.0 * dr[1];
                              f[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1];

                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force2_in_same_cell(
    vector2 **r,vector2 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector2 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector2 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
            /*  different species  */
        for(i2 = 0;i2 < i1;i2++)
        {
            epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];

                    dr2 = SQR(dr[0]) + SQR(dr[1]);

                    f[i1][j1][0] += epsilon * exp(-dr2) * 2.0 * dr[0];
                              f[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0];
f[i1][j1][1] += epsilon * exp(-dr2) * 2.0 * dr[1];
                              f[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1];


                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }

            /*  same species  */
        epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i1];

        j1 = celllist->head[i1][k];
        while(!(j1 < 0))
        {
            j2 = celllist->head[i1][k];
            while(!(j2 < 0))
            {
                if(j1 < j2)
                {
                    dr[0] = r[i1][j1][0] - r[i1][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i1][j2][1] + (*c)[1];

                    dr2 = SQR(dr[0]) + SQR(dr[1]);

                    f[i1][j1][0] += epsilon * exp(-dr2) * 2.0 * dr[0];
                              f[i1][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0];
f[i1][j1][1] += epsilon * exp(-dr2) * 2.0 * dr[1];
                              f[i1][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1];

                }

                j2 = celllist->list[i1][j2];
            }

            j1 = celllist->list[i1][j1];
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_force3(vector3 **r,vector3 **f,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    int i,j,k;
    vector3 dr;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];
dr[2] = r[i][j][2] - r[k][j][2];

                if(dr[0] >= 0.5 * geometry->l[0])
                          {
                              dr[0] -= geometry->l[0];
                          }
                          else if(dr[0] < - 0.5 * geometry->l[0])
                          {
                              dr[0] += geometry->l[0];
                          }
if(dr[1] >= 0.5 * geometry->l[1])
                          {
                              dr[1] -= geometry->l[1];
                          }
                          else if(dr[1] < - 0.5 * geometry->l[1])
                          {
                              dr[1] += geometry->l[1];
                          }
if(dr[2] >= 0.5 * geometry->l[2])
                          {
                              dr[2] -= geometry->l[2];
                          }
                          else if(dr[2] < - 0.5 * geometry->l[2])
                          {
                              dr[2] += geometry->l[2];
                          }

                f[i][j][0] -= dr[0];
                          f[k][j][0] += dr[0];
f[i][j][1] -= dr[1];
                          f[k][j][1] += dr[1];
f[i][j][2] -= dr[2];
                          f[k][j][2] += dr[2];
            }
        }
    }
    else
    {
            /*  box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];
dr[2] = r[i][j][2] - r[k][j][2];

                if(dr[1] >= 0.5 * geometry->l[1])
                {
                    dr[1] -= geometry->l[1];
                    dr[0] -= geometry->delta_l;
                }
                else if(dr[1] < - 0.5 * geometry->l[1])
                {
                    dr[1] += geometry->l[1];
                    dr[0] += geometry->delta_l;
                }
                while(dr[0] >= 0.5 * geometry->l[0])
                {
                    dr[0] -= geometry->l[0];
                }
                while(dr[0] < - 0.5 * geometry->l[0])
                {
                    dr[0] += geometry->l[0];
                }
                if(dr[2] >= 0.5 * geometry->l[2])
                       {
                           dr[2] -= geometry->l[2];
                       }
                       else if(dr[2] < - 0.5 * geometry->l[2])
                       {
                           dr[2] += geometry->l[2];
                       }

                f[i][j][0] -= dr[0];
                          f[k][j][0] += dr[0];
f[i][j][1] -= dr[1];
                          f[k][j][1] += dr[1];
f[i][j][2] -= dr[2];
                          f[k][j][2] += dr[2];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force3(vector3 **r,vector3 **f,
                                          linked_cell_list *celllist,
                                          polymer_blend *blend,
                                          box_geometry *geometry)
{
    int i1,j1,k1,i2,j2,k2,i2_i,j2_i,k2_i;
    int cell_index1,cell_index2;
    vector3 c;
    int i2_min,i2_max;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(k1 = 0;k1 < celllist->nz;k1++)
               {
        for(j1 = 0;j1 < celllist->ny;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;

                for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_force3_in_same_cell(
                                r,f,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_force3_in_different_cells(
                                r,f,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                }

            }
        }
        }
    }
    else
    {
            /*  box deformation  */
        for(k1 = 0;k1 < celllist->nz;k1++)
               {

            /*  boundary cells (j = 0, ny - 1)  */
        for(j1 = 0;j1 < celllist->ny;j1 += (celllist->ny - 1))
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;

                for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else
                    {
                        i2_min = i1 - 1;
                        i2_max = i1 + 1;
                    }

                    for(i2 = i2_min;i2 <= i2_max;i2++)
                    {
                        i2_i = i2;
                        c[0] = 0.0;

                        while(i2_i < 0)
                        {
                            i2_i += celllist->nx;
                            c[0] += geometry->l[0];
                        }
                        while(i2_i >= celllist->nx)
                        {
                            i2_i -= celllist->nx;
                            c[0] -= geometry->l[0];
                        }

                        if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] += geometry->delta_l;
                        }
                        else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] -= geometry->delta_l;
                        }


                        cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;

                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_force3_in_same_cell(
                                r,f,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_force3_in_different_cells(
                                r,f,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                }
            }
        }


            /*  non-boundary cells  */
        for(j1 = 1;j1 < celllist->ny - 1;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;

                for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_force3_in_same_cell(
                                r,f,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_force3_in_different_cells(
                                r,f,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                }

            }
        }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force3_in_different_cells(
    vector3 **r,vector3 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector3 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector3 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
        for(i2 = 0;i2 < blend->n;i2++)
        {
            epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k1];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k2];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];
dr[2] = r[i1][j1][2] - r[i2][j2][2] + (*c)[2];

                    dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);

                    f[i1][j1][0] += epsilon * exp(-dr2) * 2.0 * dr[0];
                              f[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0];
f[i1][j1][1] += epsilon * exp(-dr2) * 2.0 * dr[1];
                              f[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1];
f[i1][j1][2] += epsilon * exp(-dr2) * 2.0 * dr[2];
                              f[i2][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[2];

                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force3_in_same_cell(
    vector3 **r,vector3 **f,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector3 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector3 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
            /*  different species  */
        for(i2 = 0;i2 < i1;i2++)
        {
            epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];
dr[2] = r[i1][j1][2] - r[i2][j2][2] + (*c)[2];

                    dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);

                    f[i1][j1][0] += epsilon * exp(-dr2) * 2.0 * dr[0];
                              f[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0];
f[i1][j1][1] += epsilon * exp(-dr2) * 2.0 * dr[1];
                              f[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1];
f[i1][j1][2] += epsilon * exp(-dr2) * 2.0 * dr[2];
                              f[i2][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[2];


                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }

            /*  same species  */
        epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i1];

        j1 = celllist->head[i1][k];
        while(!(j1 < 0))
        {
            j2 = celllist->head[i1][k];
            while(!(j2 < 0))
            {
                if(j1 < j2)
                {
                    dr[0] = r[i1][j1][0] - r[i1][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i1][j2][1] + (*c)[1];
dr[2] = r[i1][j1][2] - r[i1][j2][2] + (*c)[2];

                    dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);

                    f[i1][j1][0] += epsilon * exp(-dr2) * 2.0 * dr[0];
                              f[i1][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0];
f[i1][j1][1] += epsilon * exp(-dr2) * 2.0 * dr[1];
                              f[i1][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1];
f[i1][j1][2] += epsilon * exp(-dr2) * 2.0 * dr[2];
                              f[i1][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[2];

                }

                j2 = celllist->list[i1][j2];
            }

            j1 = celllist->list[i1][j1];
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_stress2(vector2 **r,symmetric_tensor2 **sigma,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    int i,j,k;
    vector2 dr;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];

                if(dr[0] >= 0.5 * geometry->l[0])
                          {
                              dr[0] -= geometry->l[0];
                          }
                          else if(dr[0] < - 0.5 * geometry->l[0])
                          {
                              dr[0] += geometry->l[0];
                          }
if(dr[1] >= 0.5 * geometry->l[1])
                          {
                              dr[1] -= geometry->l[1];
                          }
                          else if(dr[1] < - 0.5 * geometry->l[1])
                          {
                              dr[1] += geometry->l[1];
                          }

                sigma[i][j][0] += dr[0] * dr[0];
                                    sigma[k][j][0] += dr[0] * dr[0];
sigma[i][j][1] += dr[1] * dr[1];
                                    sigma[k][j][1] += dr[1] * dr[1];
sigma[i][j][2] += dr[0] * dr[1];
                                    sigma[k][j][2] += dr[0] * dr[1];
            }
        }
    }
    else
    {
            /*  box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];

                if(dr[1] >= 0.5 * geometry->l[1])
                {
                    dr[1] -= geometry->l[1];
                    dr[0] -= geometry->delta_l;
                }
                else if(dr[1] < - 0.5 * geometry->l[1])
                {
                    dr[1] += geometry->l[1];
                    dr[0] += geometry->delta_l;
                }
                while(dr[0] >= 0.5 * geometry->l[0])
                {
                    dr[0] -= geometry->l[0];
                }
                while(dr[0] < - 0.5 * geometry->l[0])
                {
                    dr[0] += geometry->l[0];
                }
                

                sigma[i][j][0] += dr[0] * dr[0];
                                    sigma[k][j][0] += dr[0] * dr[0];
sigma[i][j][1] += dr[1] * dr[1];
                                    sigma[k][j][1] += dr[1] * dr[1];
sigma[i][j][2] += dr[0] * dr[1];
                                    sigma[k][j][2] += dr[0] * dr[1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress2(vector2 **r,symmetric_tensor2 **sigma,
                                          linked_cell_list *celllist,
                                          polymer_blend *blend,
                                          box_geometry *geometry)
{
    int i1,j1,i2,j2,i2_i,j2_i;
    int cell_index1,cell_index2;
    vector2 c;
    int i2_min,i2_max;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        
        for(j1 = 0;j1 < celllist->ny;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = j1 * celllist->nx + i1;

                
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = j2_i * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_stress2_in_same_cell(
                                r,sigma,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_stress2_in_different_cells(
                                r,sigma,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                

            }
        }
        
    }
    else
    {
            /*  box deformation  */
        

            /*  boundary cells (j = 0, ny - 1)  */
        for(j1 = 0;j1 < celllist->ny;j1 += (celllist->ny - 1))
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = j1 * celllist->nx + i1;

                
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else
                    {
                        i2_min = i1 - 1;
                        i2_max = i1 + 1;
                    }

                    for(i2 = i2_min;i2 <= i2_max;i2++)
                    {
                        i2_i = i2;
                        c[0] = 0.0;

                        while(i2_i < 0)
                        {
                            i2_i += celllist->nx;
                            c[0] += geometry->l[0];
                        }
                        while(i2_i >= celllist->nx)
                        {
                            i2_i -= celllist->nx;
                            c[0] -= geometry->l[0];
                        }

                        if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] += geometry->delta_l;
                        }
                        else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] -= geometry->delta_l;
                        }


                        cell_index2 = j2_i * celllist->nx + i2_i;

                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_stress2_in_same_cell(
                                r,sigma,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_stress2_in_different_cells(
                                r,sigma,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                
            }
        }


            /*  non-boundary cells  */
        for(j1 = 1;j1 < celllist->ny - 1;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = j1 * celllist->nx + i1;

                
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = j2_i * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_stress2_in_same_cell(
                                r,sigma,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_stress2_in_different_cells(
                                r,sigma,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                

            }
        }
        
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress2_in_different_cells(
    vector2 **r,symmetric_tensor2 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector2 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector2 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
        for(i2 = 0;i2 < blend->n;i2++)
        {
            epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k1];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k2];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];

                    dr2 = SQR(dr[0]) + SQR(dr[1]);

                    sigma[i1][j1][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
                                        sigma[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
sigma[i1][j1][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
                                        sigma[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
sigma[i1][j1][2] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
                                        sigma[i2][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];

                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress2_in_same_cell(
    vector2 **r,symmetric_tensor2 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector2 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector2 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
            /*  different species  */
        for(i2 = 0;i2 < i1;i2++)
        {
            epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];

                    dr2 = SQR(dr[0]) + SQR(dr[1]);

                    sigma[i1][j1][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
                                        sigma[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
sigma[i1][j1][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
                                        sigma[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
sigma[i1][j1][2] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
                                        sigma[i2][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];


                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }

            /*  same species  */
        epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i1];

        j1 = celllist->head[i1][k];
        while(!(j1 < 0))
        {
            j2 = celllist->head[i1][k];
            while(!(j2 < 0))
            {
                if(j1 < j2)
                {
                    dr[0] = r[i1][j1][0] - r[i1][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i1][j2][1] + (*c)[1];

                    dr2 = SQR(dr[0]) + SQR(dr[1]);

                    sigma[i1][j1][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
                                        sigma[i1][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
sigma[i1][j1][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
                                        sigma[i1][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
sigma[i1][j1][2] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
                                        sigma[i1][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];

                }

                j2 = celllist->list[i1][j2];
            }

            j1 = celllist->list[i1][j1];
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_stress3(vector3 **r,symmetric_tensor3 **sigma,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    int i,j,k;
    vector3 dr;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];
dr[2] = r[i][j][2] - r[k][j][2];

                if(dr[0] >= 0.5 * geometry->l[0])
                          {
                              dr[0] -= geometry->l[0];
                          }
                          else if(dr[0] < - 0.5 * geometry->l[0])
                          {
                              dr[0] += geometry->l[0];
                          }
if(dr[1] >= 0.5 * geometry->l[1])
                          {
                              dr[1] -= geometry->l[1];
                          }
                          else if(dr[1] < - 0.5 * geometry->l[1])
                          {
                              dr[1] += geometry->l[1];
                          }
if(dr[2] >= 0.5 * geometry->l[2])
                          {
                              dr[2] -= geometry->l[2];
                          }
                          else if(dr[2] < - 0.5 * geometry->l[2])
                          {
                              dr[2] += geometry->l[2];
                          }

                sigma[i][j][0] += dr[0] * dr[0];
                                    sigma[k][j][0] += dr[0] * dr[0];
sigma[i][j][1] += dr[1] * dr[1];
                                    sigma[k][j][1] += dr[1] * dr[1];
sigma[i][j][2] += dr[2] * dr[2];
                                    sigma[k][j][2] += dr[2] * dr[2];
sigma[i][j][3] += dr[0] * dr[1];
                                    sigma[k][j][3] += dr[0] * dr[1];
sigma[i][j][4] += dr[1] * dr[2];
                                    sigma[k][j][4] += dr[1] * dr[2];
sigma[i][j][5] += dr[2] * dr[0];
                                    sigma[k][j][5] += dr[2] * dr[0];
            }
        }
    }
    else
    {
            /*  box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                dr[0] = r[i][j][0] - r[k][j][0];
dr[1] = r[i][j][1] - r[k][j][1];
dr[2] = r[i][j][2] - r[k][j][2];

                if(dr[1] >= 0.5 * geometry->l[1])
                {
                    dr[1] -= geometry->l[1];
                    dr[0] -= geometry->delta_l;
                }
                else if(dr[1] < - 0.5 * geometry->l[1])
                {
                    dr[1] += geometry->l[1];
                    dr[0] += geometry->delta_l;
                }
                while(dr[0] >= 0.5 * geometry->l[0])
                {
                    dr[0] -= geometry->l[0];
                }
                while(dr[0] < - 0.5 * geometry->l[0])
                {
                    dr[0] += geometry->l[0];
                }
                if(dr[2] >= 0.5 * geometry->l[2])
                       {
                           dr[2] -= geometry->l[2];
                       }
                       else if(dr[2] < - 0.5 * geometry->l[2])
                       {
                           dr[2] += geometry->l[2];
                       }

                sigma[i][j][0] += dr[0] * dr[0];
                                    sigma[k][j][0] += dr[0] * dr[0];
sigma[i][j][1] += dr[1] * dr[1];
                                    sigma[k][j][1] += dr[1] * dr[1];
sigma[i][j][2] += dr[2] * dr[2];
                                    sigma[k][j][2] += dr[2] * dr[2];
sigma[i][j][3] += dr[0] * dr[1];
                                    sigma[k][j][3] += dr[0] * dr[1];
sigma[i][j][4] += dr[1] * dr[2];
                                    sigma[k][j][4] += dr[1] * dr[2];
sigma[i][j][5] += dr[2] * dr[0];
                                    sigma[k][j][5] += dr[2] * dr[0];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress3(vector3 **r,symmetric_tensor3 **sigma,
                                          linked_cell_list *celllist,
                                          polymer_blend *blend,
                                          box_geometry *geometry)
{
    int i1,j1,k1,i2,j2,k2,i2_i,j2_i,k2_i;
    int cell_index1,cell_index2;
    vector3 c;
    int i2_min,i2_max;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(k1 = 0;k1 < celllist->nz;k1++)
               {
        for(j1 = 0;j1 < celllist->ny;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;

                for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_stress3_in_same_cell(
                                r,sigma,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_stress3_in_different_cells(
                                r,sigma,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                }

            }
        }
        }
    }
    else
    {
            /*  box deformation  */
        for(k1 = 0;k1 < celllist->nz;k1++)
               {

            /*  boundary cells (j = 0, ny - 1)  */
        for(j1 = 0;j1 < celllist->ny;j1 += (celllist->ny - 1))
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;

                for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else
                    {
                        i2_min = i1 - 1;
                        i2_max = i1 + 1;
                    }

                    for(i2 = i2_min;i2 <= i2_max;i2++)
                    {
                        i2_i = i2;
                        c[0] = 0.0;

                        while(i2_i < 0)
                        {
                            i2_i += celllist->nx;
                            c[0] += geometry->l[0];
                        }
                        while(i2_i >= celllist->nx)
                        {
                            i2_i -= celllist->nx;
                            c[0] -= geometry->l[0];
                        }

                        if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] += geometry->delta_l;
                        }
                        else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] -= geometry->delta_l;
                        }


                        cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;

                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_stress3_in_same_cell(
                                r,sigma,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_stress3_in_different_cells(
                                r,sigma,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                }
            }
        }


            /*  non-boundary cells  */
        for(j1 = 1;j1 < celllist->ny - 1;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;

                for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_stress3_in_same_cell(
                                r,sigma,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_stress3_in_different_cells(
                                r,sigma,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                }

            }
        }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress3_in_different_cells(
    vector3 **r,symmetric_tensor3 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector3 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector3 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
        for(i2 = 0;i2 < blend->n;i2++)
        {
            epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k1];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k2];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];
dr[2] = r[i1][j1][2] - r[i2][j2][2] + (*c)[2];

                    dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);

                    sigma[i1][j1][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
                                        sigma[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
sigma[i1][j1][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
                                        sigma[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
sigma[i1][j1][2] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[2];
                                        sigma[i2][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[2];
sigma[i1][j1][3] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
                                        sigma[i2][j2][3] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
sigma[i1][j1][4] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[2];
                                        sigma[i2][j2][4] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[2];
sigma[i1][j1][5] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[0];
                                        sigma[i2][j2][5] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[0];

                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress3_in_same_cell(
    vector3 **r,symmetric_tensor3 **sigma,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector3 *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector3 dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
            /*  different species  */
        for(i2 = 0;i2 < i1;i2++)
        {
            epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i2];

            j1 = celllist->head[i1][k];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k];
                while(!(j2 < 0))
                {
                    dr[0] = r[i1][j1][0] - r[i2][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i2][j2][1] + (*c)[1];
dr[2] = r[i1][j1][2] - r[i2][j2][2] + (*c)[2];

                    dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);

                    sigma[i1][j1][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
                                        sigma[i2][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
sigma[i1][j1][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
                                        sigma[i2][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
sigma[i1][j1][2] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[2];
                                        sigma[i2][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[2];
sigma[i1][j1][3] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
                                        sigma[i2][j2][3] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
sigma[i1][j1][4] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[2];
                                        sigma[i2][j2][4] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[2];
sigma[i1][j1][5] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[0];
                                        sigma[i2][j2][5] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[0];


                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }

            /*  same species  */
        epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i1];

        j1 = celllist->head[i1][k];
        while(!(j1 < 0))
        {
            j2 = celllist->head[i1][k];
            while(!(j2 < 0))
            {
                if(j1 < j2)
                {
                    dr[0] = r[i1][j1][0] - r[i1][j2][0] + (*c)[0];
dr[1] = r[i1][j1][1] - r[i1][j2][1] + (*c)[1];
dr[2] = r[i1][j1][2] - r[i1][j2][2] + (*c)[2];

                    dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);

                    sigma[i1][j1][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
                                        sigma[i1][j2][0] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[0];
sigma[i1][j1][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
                                        sigma[i1][j2][1] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[1];
sigma[i1][j1][2] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[2];
                                        sigma[i1][j2][2] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[2];
sigma[i1][j1][3] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
                                        sigma[i1][j2][3] -= epsilon * exp(-dr2) * 2.0 * dr[0] * dr[1];
sigma[i1][j1][4] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[2];
                                        sigma[i1][j2][4] -= epsilon * exp(-dr2) * 2.0 * dr[1] * dr[2];
sigma[i1][j1][5] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[0];
                                        sigma[i1][j2][5] -= epsilon * exp(-dr2) * 2.0 * dr[2] * dr[0];

                }

                j2 = celllist->list[i1][j2];
            }

            j1 = celllist->list[i1][j1];
        }
    }
}

/*---------------------------------------------------------------------------*/
